# -*- coding: iso-8859-1 -*-
"""
This module implements TomTom's compression methods
"""

__version__ = '$Id$'

import struct

BLOCK_SIZE = 8192 # read x bytes at once from file

LOOKBACK   = 4096 # lookback depth for tomtom compress
MAXLENGTH  = 17   # lookback length for tomtom compress
                  # (18 is better, but tomtom uses 17)


#
# Read uncompressed item (0x00)
#
def uncompress_plain(inp, out, osize, iend, callback=None):
	"""
	Uncompress plain data (ie. don't uncompress at all ;-)

	@param inp: input handle
	@type  inp: Filehandle
	@param out: output handle
	@type  out: Filehandle

	@param osize: output size to be filled
	@type  osize: Int
	@param iend: end of input stream at pos
	@type  iend: Int

	@param callback: progress indicator callback function
	@type  callback: Function or None

	@return: number of input bytes read
	@rtype : Int
	"""

	istart = inp.tell()
	osize = min(osize, iend)

	if callback is not None:
		callback(0, osize)

	left = osize
	while left:
		size = min(left, BLOCK_SIZE)
		data = inp.read(size)
		if len(data) != size:
			raise IOError('short read for uncompressed item')
		else:
			left -= len(data)
			out.write(data)
			if callback is not None:
				callback(osize-left)

	return inp.tell() - istart

#
# Read compressed item (0x04)
#
def uncompress_tomtom(inp, out, osize, iend, callback=None):
	"""
	Uncompress tomtom compressed data

	@param inp: input handle
	@type  inp: Filehandle
	@param out: output handle
	@type  out: Filehandle

	@param osize: output size to be filled
	@type  osize: Int
	@param iend: end of input stream at pos
	@type  iend: Int

	@param callback: progress indicator callback function
	@type  callback: Function or None

	@return: number of input bytes read
	@rtype : Int
	"""

	istart = inp.tell()
	ostart = out.tell()

	if callback is not None:
		callback(0, osize)

	while out.tell()-ostart < osize:
		try:
			if inp.tell() >= iend:
				blockhead = 0
			else:
				blockhead = ord(inp.read(1))
		except TypeError:
			raise IOError('short read for compression head')
		else:
			for cnt in range(0, 8):
				if blockhead & 1:
					try:
						if inp.tell() >= iend:
							item = 0
						else:
							item, = struct.unpack('>H', inp.read(2))
					except struct.error:
						raise IOError('short read for compressed block at bit %d' % (cnt))
					else:
						offset = 1 + (item & 0x0fff)
						length = 3 + (item >> 12)
						if offset > out.tell()-ostart:
							raise IOError('invalid offset in compressed block at bit %d' % \
							               (cnt))
						elif offset <= length:
							out.seek(-offset, 1)
							out.write((out.read(offset)*((length-1)/offset+1))[:length])
						else:
							out.seek(-offset, 1)
							add = out.read(length)
							out.seek(offset-length, 1)
							out.write(add)
				elif out.tell()-ostart < osize:
					if inp.tell() >= iend:
						item = '\0'
					else:
						item = inp.read(1)
					if len(item) != 1:
						raise IOError('short read for compressed block at bit %d' % (cnt))
					else:
						out.write(item)
				blockhead >>= 1
				if callback is not None:
					callback(out.tell()-ostart)

	return inp.tell() - istart

#
# Write uncompressed item (0x00)
#
def compress_plain(inp, out, isize, callback=None):
	"""
	Compress plain data (ie. don't compress at all ;-)

	@param inp: input handle
	@type  inp: Filehandle
	@param out: output handle
	@type  out: Filehandle

	@param isize: input size to be read
	@type  isize: Int
	@param callback: progress indicator callback function
	@type  callback: Function or None

	@return: number of output bytes written
	@rtype : Int
	"""

	ostart = out.tell()

	if callback is not None:
		callback(0, isize)

	left = isize
	while left:
		size = min(left, BLOCK_SIZE)
		data = inp.read(size)
		if len(data) != size:
			raise IOError('short read for input data')
		else:
			left -= len(data)
			out.write(data)
			if callback is not None:
				callback(isize-left)

	out.write('\0'*(3-(out.tell()-ostart-1)%4))

	return out.tell() - ostart

#
# Write compressed item (0x04)
#
def compress_tomtom(inp, out, isize, callback=None):
	"""
	Compress tomtom compressed data

	@param inp: input handle
	@type  inp: Filehandle
	@param out: output handle
	@type  out: Filehandle

	@param isize: input size to be read
	@type  isize: Int
	@param callback: progress indicator callback function
	@type  callback: Function or None

	@return: number of output bytes written
	@rtype : Int
	"""

	ostart = out.tell()

	if callback is not None:
		callback(0, isize)

	idata = inp.read(isize)
	if len(idata) != isize:
		raise IOError('short read for input data')

	ipos  = 0
	istop = isize-2
	obit  = 1
	obyte = 0
	odata = ''

	while ipos < isize:

		if obit == 256:
			out.write(chr(obyte))
			out.write(odata)
			obit  = 1
			obyte = 0
			odata = ''
			if callback is not None:
				callback(ipos)

		if 0 < ipos < istop:

			dseek = idata[ipos:ipos+3]
			last_found  = None
			last_length = 2

			look_from = max(0, ipos-LOOKBACK)
			look_to   = ipos+2
			look_stop = min(isize-ipos-1, MAXLENGTH)
			while look_to:
				found = idata.rfind(dseek, look_from, look_to)
				if found == -1:
					break
				else:
					length = 1
					while length < look_stop:
						if idata[found+length] != idata[ipos+length]:
							break
						else:
							length += 1
					if length > last_length:
						last_found  = found
						last_length = length
						if length == MAXLENGTH or ipos+length == isize:
							break
					look_to = found+2

			if last_found is None:
				odata += dseek[0]
				ipos += 1
			else:
				obyte |= obit
				odata += struct.pack('>H', ((last_length-3)<<12)+ipos-last_found-1)
				ipos += last_length

		else:
			odata += idata[ipos]
			ipos += 1

		obit <<= 1

	if obit > 1:
		out.write(chr(obyte))
		out.write(odata)
		if obit == 256:
			out.write('\0')
		if callback is not None:
			callback(ipos)
	else:
		out.write('\0')

	return out.tell() - ostart

